/* getProductName.cpp
 *   derived from:
 *     Microsoft example code "Getting the System Version"
 *     strlcat, strlcpy: OpenBSD [BSD license]
 *
 * Modifications Copyright (c) 2008, Charles S. Wilson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* must include before windows.h */
#include "getopt.h"

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include "../version.h"

static const char* VERSION = VERSION_STRING;
static const char* PROGNAME= "winProductName";

/* w32api does not yet define these constants */
#define PRODUCT_BUSINESS				0x00000006
#define PRODUCT_BUSINESS_N				0x00000010
#define PRODUCT_CLUSTER_SERVER				0x00000012
#define PRODUCT_DATACENTER_SERVER			0x00000008
#define PRODUCT_DATACENTER_SERVER_CORE			0x0000000C
#define PRODUCT_DATACENTER_SERVER_CORE_V		0x00000027
#define PRODUCT_DATACENTER_SERVER_V			0x00000025
#define PRODUCT_ENTERPRISE				0x00000004
#define PRODUCT_ENTERPRISE_N				0x0000001B
#define PRODUCT_ENTERPRISE_SERVER			0x0000000A
#define PRODUCT_ENTERPRISE_SERVER_CORE			0x0000000E
#define PRODUCT_ENTERPRISE_SERVER_CORE_V		0x00000029
#define PRODUCT_ENTERPRISE_SERVER_IA64			0x0000000F
#define PRODUCT_ENTERPRISE_SERVER_V			0x00000026
#define PRODUCT_HOME_BASIC				0x00000002
#define PRODUCT_HOME_BASIC_N				0x00000005
#define PRODUCT_HOME_PREMIUM				0x00000003
#define PRODUCT_HOME_PREMIUM_N				0x0000001A
#define PRODUCT_HOME_SERVER				0x00000013
#define PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT	0x0000001E
#define PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING		0x00000020
#define PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY		0x0000001F
#define PRODUCT_SERVER_FOR_SMALLBUSINESS		0x00000018
#define PRODUCT_SMALLBUSINESS_SERVER			0x00000009
#define PRODUCT_SMALLBUSINESS_SERVER_PREMIUM		0x00000019
#define PRODUCT_STANDARD_SERVER				0x00000007
#define PRODUCT_STANDARD_SERVER_CORE			0x0000000D
#define PRODUCT_STANDARD_SERVER_CORE_V			0x00000028
#define PRODUCT_STANDARD_SERVER_V			0x00000024
#define PRODUCT_STARTER					0x0000000B
#define PRODUCT_STORAGE_ENTERPRISE_SERVER		0x00000017
#define PRODUCT_STORAGE_EXPRESS_SERVER			0x00000014
#define PRODUCT_STORAGE_STANDARD_SERVER			0x00000015
#define PRODUCT_STORAGE_WORKGROUP_SERVER		0x00000016
#define PRODUCT_UNDEFINED				0x00000000
#define PRODUCT_ULTIMATE				0x00000001
#define PRODUCT_ULTIMATE_N				0x0000001C
#define PRODUCT_WEB_SERVER				0x00000011
#define PRODUCT_WEB_SERVER_CORE				0x0000001D


#define BUFSIZE 256

static size_t strlcpy(char *dst, const char *src, size_t siz);
static size_t strlcat(char *dst, const char *src, size_t siz);
void usage  (FILE* os, char* progname, int exitCode);
void help   (FILE* os, char* progname, int exitCode);
void license(FILE* os, char* progname, int exitCode);
void version(FILE* os, char* progname, int exitCode);

typedef void (WINAPI *PGNSI)(LPSYSTEM_INFO);
typedef BOOL (WINAPI *PGPI)(DWORD, DWORD, DWORD, DWORD, PDWORD);

BOOL GetOSDisplayString( LPTSTR pszOS)
{
   OSVERSIONINFOEX osvi;
   SYSTEM_INFO si;
   PGNSI pGNSI;
   PGPI pGPI;
   BOOL bOsVersionInfoEx;
   DWORD dwType;

   ZeroMemory(&si, sizeof(SYSTEM_INFO));
   ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));

   osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

   if( !(bOsVersionInfoEx = GetVersionEx ((OSVERSIONINFO *) &osvi)) )
      return 1;

   // Call GetNativeSystemInfo if supported or GetSystemInfo otherwise.

   pGNSI = (PGNSI) GetProcAddress(
      GetModuleHandle(TEXT("kernel32.dll")),
      "GetNativeSystemInfo");
   if(NULL != pGNSI)
      pGNSI(&si);
   else GetSystemInfo(&si);

   if ( VER_PLATFORM_WIN32_NT==osvi.dwPlatformId &&
        osvi.dwMajorVersion > 4 )
   {
      strlcpy(pszOS, TEXT("Microsoft "), BUFSIZE);

      // Test for the specific product.
      if ( osvi.dwMajorVersion == 6 && osvi.dwMinorVersion == 1 )
      {
         if( osvi.wProductType == VER_NT_WORKSTATION )
             strlcat(pszOS, TEXT("Windows 7 "), BUFSIZE);
         else strlcat(pszOS, TEXT("Windows Server 2008 R2 " ), BUFSIZE);
      }

      if ( osvi.dwMajorVersion == 6 && osvi.dwMinorVersion == 0 )
      {
         if( osvi.wProductType == VER_NT_WORKSTATION )
             strlcat(pszOS, TEXT("Windows Vista "), BUFSIZE);
         else strlcat(pszOS, TEXT("Windows Server 2008 " ), BUFSIZE);

         pGPI = (PGPI) GetProcAddress(
            GetModuleHandle(TEXT("kernel32.dll")),
            "GetProductInfo");

         pGPI( 6, 0, 0, 0, &dwType);

         switch( dwType )
         {
            case PRODUCT_ULTIMATE:
               strlcat (pszOS, TEXT("Ultimate Edition" ), BUFSIZE);
               break;
            case PRODUCT_HOME_PREMIUM:
               strlcat (pszOS, TEXT("Home Premium Edition" ), BUFSIZE);
               break;
            case PRODUCT_HOME_BASIC:
               strlcat (pszOS, TEXT("Home Basic Edition" ), BUFSIZE);
               break;
            case PRODUCT_ENTERPRISE:
               strlcat (pszOS, TEXT("Enterprise Edition" ), BUFSIZE);
               break;
            case PRODUCT_BUSINESS:
               strlcat (pszOS, TEXT("Business Edition" ), BUFSIZE);
               break;
            case PRODUCT_STARTER:
               strlcat (pszOS, TEXT("Starter Edition" ), BUFSIZE);
               break;
            case PRODUCT_CLUSTER_SERVER:
               strlcat (pszOS, TEXT("Cluster Server Edition" ), BUFSIZE);
               break;
            case PRODUCT_DATACENTER_SERVER:
               strlcat (pszOS, TEXT("Datacenter Edition" ), BUFSIZE);
               break;
            case PRODUCT_DATACENTER_SERVER_CORE:
               strlcat (pszOS, TEXT("Datacenter Edition (core installation)" ), BUFSIZE);
               break;
            case PRODUCT_ENTERPRISE_SERVER:
               strlcat (pszOS, TEXT("Enterprise Edition" ), BUFSIZE);
               break;
            case PRODUCT_ENTERPRISE_SERVER_CORE:
               strlcat (pszOS, TEXT("Enterprise Edition (core installation)" ), BUFSIZE);
               break;
            case PRODUCT_ENTERPRISE_SERVER_IA64:
               strlcat (pszOS, TEXT("Enterprise Edition for Itanium-based Systems" ), BUFSIZE);
               break;
            case PRODUCT_SMALLBUSINESS_SERVER:
               strlcat (pszOS, TEXT("Small Business Server" ), BUFSIZE);
               break;
            case PRODUCT_SMALLBUSINESS_SERVER_PREMIUM:
               strlcat (pszOS, TEXT("Small Business Server Premium Edition" ), BUFSIZE);
               break;
            case PRODUCT_STANDARD_SERVER:
               strlcat (pszOS, TEXT("Standard Edition" ), BUFSIZE);
               break;
            case PRODUCT_STANDARD_SERVER_CORE:
               strlcat (pszOS, TEXT("Standard Edition (core installation)" ), BUFSIZE);
               break;
            case PRODUCT_WEB_SERVER:
               strlcat (pszOS, TEXT("Web Server Edition" ), BUFSIZE);
               break;
         }
         if ( si.wProcessorArchitecture==PROCESSOR_ARCHITECTURE_AMD64 )
            strlcat (pszOS, TEXT( ", 64-bit" ), BUFSIZE);
         else if (si.wProcessorArchitecture==PROCESSOR_ARCHITECTURE_INTEL )
            strlcat (pszOS, TEXT(", 32-bit"), BUFSIZE);
      }

      if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2 )
      {
         if( GetSystemMetrics(SM_SERVERR2) )
            strlcat (pszOS, TEXT( "Windows Server 2003 R2, "), BUFSIZE);
         else if ( osvi.wSuiteMask==VER_SUITE_STORAGE_SERVER )
            strlcat (pszOS, TEXT( "Windows Storage Server 2003"), BUFSIZE);
         else if( osvi.wProductType == VER_NT_WORKSTATION &&
                  si.wProcessorArchitecture==PROCESSOR_ARCHITECTURE_AMD64)
         {
            strlcat (pszOS, TEXT( "Windows XP Professional x64 Edition"), BUFSIZE);
         }
         else strlcat (pszOS, TEXT("Windows Server 2003, "), BUFSIZE);

         // Test for the server type.
         if ( osvi.wProductType != VER_NT_WORKSTATION )
         {
            if ( si.wProcessorArchitecture==PROCESSOR_ARCHITECTURE_IA64 )
            {
                if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
                   strlcat (pszOS, TEXT( "Datacenter Edition for Itanium-based Systems" ), BUFSIZE);
                else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
                   strlcat (pszOS, TEXT( "Enterprise Edition for Itanium-based Systems" ), BUFSIZE);
            }

            else if ( si.wProcessorArchitecture==PROCESSOR_ARCHITECTURE_AMD64 )
            {
                if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
                   strlcat (pszOS, TEXT( "Datacenter x64 Edition" ), BUFSIZE);
                else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
                   strlcat (pszOS, TEXT( "Enterprise x64 Edition" ), BUFSIZE);
                else strlcat (pszOS, TEXT( "Standard x64 Edition" ), BUFSIZE);
            }

            else
            {
                if ( osvi.wSuiteMask & VER_SUITE_COMPUTE_SERVER )
                   strlcat (pszOS, TEXT( "Compute Cluster Edition" ), BUFSIZE);
                else if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
                   strlcat (pszOS, TEXT( "Datacenter Edition" ), BUFSIZE);
                else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
                   strlcat (pszOS, TEXT( "Enterprise Edition" ), BUFSIZE);
                else if ( osvi.wSuiteMask & VER_SUITE_BLADE )
                   strlcat (pszOS, TEXT( "Web Edition" ), BUFSIZE);
                else strlcat (pszOS, TEXT( "Standard Edition" ), BUFSIZE);
            }
         }
      }

      if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1 )
      {
         strlcat (pszOS, TEXT("Windows XP "), BUFSIZE);
         if( osvi.wSuiteMask & VER_SUITE_PERSONAL )
            strlcat (pszOS, TEXT( "Home Edition" ), BUFSIZE);
         else strlcat (pszOS, TEXT( "Professional" ), BUFSIZE);
      }

      if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 )
      {
         strlcat (pszOS, TEXT("Windows 2000 "), BUFSIZE);

         if ( osvi.wProductType == VER_NT_WORKSTATION )
         {
            strlcat (pszOS, TEXT( "Professional" ), BUFSIZE);
         }
         else 
         {
            if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
               strlcat (pszOS, TEXT( "Datacenter Server" ), BUFSIZE);
            else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
               strlcat (pszOS, TEXT( "Advanced Server" ), BUFSIZE);
            else strlcat (pszOS, TEXT( "Server" ), BUFSIZE);
         }
      }

       // Include service pack (if any) and build number.

      if( _tcslen(osvi.szCSDVersion) > 0 )
      {
          strlcat (pszOS, TEXT(" "), BUFSIZE);
          strlcat (pszOS, osvi.szCSDVersion, BUFSIZE);
      }

      TCHAR buf[80];

      snprintf ( buf, 80, TEXT(" (build %d)"), osvi.dwBuildNumber);
      strlcat (pszOS, buf, BUFSIZE);

      return TRUE;
   }

   else
   {
      printf( "This sample does not support this version of Windows.\n");
      return FALSE;
   }
}

int main(int argc, char** argv)
{
  TCHAR szOS[BUFSIZE];
  int c;

  opterr = 0;
  while (1)
  {
    static struct option long_options[] =
    {
      {"help",    no_argument, 0, 'h'},
      {"version", no_argument, 0, 'V'},
      {"license", no_argument, 0, 'L'},
      {0, 0, 0, 0}
    };
    int option_index = 0;
    c = getopt_long(argc, argv, ":hVL", long_options, &option_index);

    if (c == -1)
      break;

    switch (c)
    {
      case 'h':
        help(stdout, argv[0], 0);
        /* not reached */
        break;
      case 'V':
        version(stdout, argv[0], 0);
        /* not reached */
        break;
      case 'L':
        license(stdout, argv[0], 0);
        /* not reached */
        break;

      case ':':
        fprintf(stderr, "option -%c missing required arguments\n", optopt);
        usage (stderr, argv[0], 1);
        /* not reached */
        break;

      case '?':
        {
          if (isprint (optopt))
            fprintf (stderr, "Unknown option `-%c'.\n", optopt);
          else
            fprintf (stderr, "Unknown option character `\\x%x'.\n", optopt);
          usage (stderr, argv[0], 1);
        }
        /* not reached */
        break;

      default:
        fprintf(stderr, "internal getopt error\n");
        exit(1);
    } /* switch */
  } /* while */

  if (optind < argc)
  {
    fprintf(stderr, "additional arguments ignored: ");
    while (optind < argc)
    {
      fprintf(stderr, "\"%s", argv[optind++]);
      if (optind < argc)
        fprintf(stderr, "\", ");
      else
        fprintf(stderr, "\"\n");
    }
    usage(stderr, argv[0], 1);
    /* not reached */
  }

  if( GetOSDisplayString( szOS ) )
  {
    _tprintf( TEXT("%s\n"), szOS );
    return 0;
  }
  /* failure */
  return 1;
}



/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 * Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 */
size_t
strlcat(char *dst, const char *src, size_t siz)
{
	char *d = dst;
	const char *s = src;
	size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t
strlcpy(char *dst, const char *src, size_t siz)
{
	char *d = dst;
	const char *s = src;
	size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

void usage  (FILE* os, char* progname, int exitCode)
{
  const char* name = (progname ? progname : PROGNAME);
  FILE* s    = (os ? os : stderr);
  fprintf(s, "Usage: %s [-hVL]\n", name);
  exit(exitCode);
}

void help   (FILE* os, char* progname, int exitCode)
{
  static const char* helpText =
"returns the full, official name of the operating system\n"
"\n"
"   --help|-h           print this help\n"
"   --version|-V        print version information\n"
"   --license|-L        print license information\n";

  const char* name = (progname ? progname : PROGNAME);
  FILE* s    = (os ? os : stderr);

  fprintf(s, "%s [-hVL]\n", name);
  fprintf(s, "%s\n", helpText);
  exit(exitCode);
}

void license(FILE* os, char* progname, int exitCode)
{
  static const char* licenseText =
"Copyright (c) 2008, Charles S. Wilson\n"
"\n"
"Permission is hereby granted, free of charge, to any person obtaining a copy\n"
"of this software and associated documentation files (the \"Software\"), to deal\n"
"in the Software without restriction, including without limitation the rights\n"
"to use, copy, modify, merge, publish, distribute, sublicense, and/or sell\n"
"copies of the Software, and to permit persons to whom the Software is\n"
"furnished to do so, subject to the following conditions:\n"
"\n"
"The above copyright notice and this permission notice shall be included in\n"
"all copies or substantial portions of the Software.\n"
"\n"
"THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n"
"IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n"
"FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\n"
"AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\n"
"LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\n"
"OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\n"
"THE SOFTWARE.\n"
"\n"
"Portions of this software are provided under independent licensing terms:\n"
"getopt:      (c) 1989--2001 Free Software Foundation (GNU C Library), LGPLv2";

  const char* name = (progname ? progname : PROGNAME);
  FILE* s    = (os ? os : stderr);

  fprintf(s, "%s version %s\n", name, VERSION);
  fprintf(s, "%s\n", licenseText);
  exit(exitCode);
}
void version(FILE* os, char* progname, int exitCode)
{
  const char* name = (progname ? progname : PROGNAME);
  FILE* s    = (os ? os : stderr);
  fprintf(s, "%s version %s\n", name, VERSION);
  exit(exitCode);
}

